function [X1,X2,res]=EKSM_genLyap_LRcomm(A,LA,UA,N,C,starting_block,maxit,tol,tolY)
% function [X1,X2,res,rankX]=EKSM_genLyap_LRcomm(A,LA,UA,D,uD,rhs,maxit,tol,tolY,alpha)
% Extended Krylov subspace method for generalized Lyapunov equation with
% low-rank commuting coefficients.
%
% We solve
%
% A X + X A' + \sum_{i=1}^m N_i*X*N_i^T = C*C^T
%
% assuming the Lyapunov part being the dominant one and 
%
% [A,N_i]=A*N_i-N_i*A=U*\tilde U^T for all i=1,...,m,
%
% where U, \tilde U are all tall matrices.
%
% Input:
% A = LA*UA;   LU factorization of A
% N: cell that contains the m matrices N_i, N{i}=N_i
% starting_block: low-rank matrix we use to build the space:
%                 EK_k^\square(A,starting_block)
% C: low-rank factor of the right-hand side
% maxit = max space dimension
% tol = max final accuracy (in terms of relative residual)
% tolY = at convergence a truncated SVD up to tolY is performed in order to further reduce the rank of the 
%         computed solution 
%
% Output:
% X1,X2: solution factors   X \approx X1*X2^T
% res:  history of residual norms
%
% REMARK: the Matlab Control Toolbox, and in particular the function
% lyap.m, is requested.
%
% Please contact D. Palitta for any problem you may encouter when running the code.
%
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
% IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
% FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
% COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
% IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
% CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

tic;

normC=norm(C,'fro')^2;  

% size of the "real" rhs
ss1=size(C,2);      
[n,sh]=size(starting_block);
sA=2*sh;
m=length(N);

% if A is symmetric the Lanczos method can be employed in the basis
% construction setting kA_max=3, but the Arnoldi procedure is much more
% stable
kA_max=maxit;


% Sequence in A
starting_blockA=UA\(LA\starting_block);  
[U(1:n,1:sA),beta]=qr([starting_block,starting_blockA],0);
gamma1=beta(1:ss1,1:ss1);
ibeta=inv(beta);
beta2=gamma1*gamma1';

% Preallocation
H=zeros((maxit+1)*sA,maxit*sA); L=zeros((maxit+1)*sA,maxit*sA);
T=zeros(maxit*sA,maxit*sA);
res=zeros(maxit+2,1);
TTN=cell(m);
TTN_old=cell(m);
odds=[];

for j=1:maxit,

    jms=(j-1)*sA+1;j1s=(j+1)*sA;js=j*sA;js1=js+1; jsh=(j-1)*sA+sh;
   
    % Sequence in A
    Up(1:n,1:sh)   = A*U(1:n,jms:jsh); 
    Up(1:n,sh+1:sA) = UA\(LA\U(1:n,jsh+1:js));
    
    %new bases block (modified gram) orthogonalization
    for l=1:2
        k_min=max(1,j-kA_max);
        for kk=k_min:j
            k1=(kk-1)*sA+1; k2=kk*sA;
            coef= U(1:n,k1:k2)'*Up;
            H(k1:k2,jms:js) = H(k1:k2,jms:js)+coef; 
            Up = Up - U(:,k1:k2)*coef; 
        end
    end
    
    % normalization
    if (j<=maxit)
       [U(1:n,js1:j1s),H(js1:j1s,jms:js)]=qr(Up,0);
       hinv=inv(H(js1:j1s,jms:js));
    end
        
    I=speye(js+sA);
    if (j==1),
      L(1:j*sA+sh,(j-1)*sh+1:j*sh) =...
      [ H(1:sA+sh,1:sh)/ibeta(1:sh,1:sh), speye(sA+sh,sh)/ibeta(1:sh,1:sh)]*ibeta(1:sA,sh+1:sA);
    else
      L(1:j*sA+sA,(j-1)*sh+1:j*sh) = L(1:j*sA+sA,(j-1)*sh+1:j*sh) + H(1:j*sA+sA,jms:jms-1+sh)*rho;
    end
    odds  = [odds, jms:(jms-1+sh)];   % store the odd block columns
    evens = 1:js; evens(odds)=[];
    T(1:js+sA,odds)=H(1:js+sA,odds);   %odd columns

    T(1:js+sh,evens)=L(1:js+sh,1:j*sh);   %even columns
    L(1:j*sA+sA,j*sh+1:(j+1)*sh) = ...
       ( I(1:j*sA+sA,(js-sh+1):js)- T(1:js+sA,1:js)*H(1:js,js-sh+1:js))*hinv(sh+1:sA,sh+1:sA);
    rho = hinv(1:sh,1:sh)\hinv(1:sh,sh+1:sA);
    

    k=j;

    % Solve reduced multiterm eqn avoiding explicit projections
    for i=1:m
        % compute the projection of N{i} onto new basis blocks
        vec1=U(1:n,1:js)'*(N{i}*U(1:n,jms:js));
        vec2=-vec1';
        % assemble with the old projected N_i
        TTN_temp=zeros(js);
        TTN_temp(1:jms-1,1:jms-1)=TTN_old{i};
        TTN_temp(:,jms:js)=vec1;
        TTN_temp(jms:js,1:jms-1)=vec2(:,1:jms-1);
        TTN{i}=TTN_temp;
        TTN_old{i}=TTN{i};
        
    end
    
    % Solve the projected problem
    Y=GenSylv_smallscale(T(1:js,1:js),T(1:js,1:js)',TTN,TTN,eye(k*sA,ss1)*beta2*eye(k*sA,ss1)',1e-14,100);
    
    % compute the residual norm
    cc  = [H(js1:j1s,js-sA+1:js-sh), L(js1:j1s,(j-1)*sh+1:j*sh)];
    res(k)=sqrt(2)*norm(cc*Y(js-sA+1:js,:),'fro')/normC;
    
if (res(k)<tol), break, end
end

% Reduce rank of solution, if needed
[uY,sY,vY] = svd(Y); 
[sY,id] = sort(diag(sY));
sY = flipud(sY);
uY = uY(:,id(end:-1:1));
vY = vY(:,id(end:-1:1));
is = sum(abs(sY)>tolY);
Y1 = uY(:,1:is)*diag(sqrt(sY(1:is))); 
Y2 = vY(:,1:is)*diag(sqrt(sY(1:is)));

X1=U(1:n,1:js)*Y1; X2=U(1:n,1:js)*Y2;

KPIKtotal_time=toc;


fprintf('Its: %d Space dim %d Solution rank %d  time %d \n',k,js,is,KPIKtotal_time);
res=res(1:k);

end

