% MIMO EXAMPLE: we are solving the generalized Lyapunov equation 
%
% A*X+X*A^T+\gamma^2*\sum_{i=1}^2 N_i*X*N_i^T= C*C^T
%
% by the extended block Krylov subspace method (EKSM) where the initial
% block is selected exploiting the low-rank commutation property of the
% coefficients.
%
% Elias Jarlebring, Giampaolo Mele, Davide Palitta and Emil Ringh
% Krylov methods for low-rank commuting generalized Sylvester equations
% April 2017, ArXiv: 1704.02167
%
% DATA:
% The coeffcient matrices come from from Example 2 of Lin, Bao, Wei, "Order reduction of
% bilinear MIMO dynamical systems using new block Krylov subspaces", 2009.
% The matrix C is a nX2 random matrix which is then normalized and \gamma>0.
%
% REMARK: the Matlab Control Toolbox, and in particular the function
% lyap.m, is requested.
%
% Please contact D. Palitta for any problem you may encouter when running the code.
%
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
% IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
% FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
% COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
% IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
% CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

close all
clear all
n=1000;

% Build the matrices from Example 2 of Lin, Bao, Wei, "Order reduction of
% bilinear MIMO dynamical systems using new block Krylov subspaces", 2009
e = ones(n,1);
A = spdiags([2*e -5*e 2*e], -1:1, n, n);
N = spdiags([3*e -0*e -3*e], -1:1, n, n);
I=speye(n);
C=randn(n,2);

% Define the low-rank factor of the commutator
U=2*sqrt(3)*[I(:,1),I(:,end)];

% Run the example for different values of gamma
gamma_vect=[1/6,1/5,1/4];

% Define thresholds and precompute the LU factors of A
maxit=100;
tol=1e-6;
tolY=1e-12;
[LA,UA]=lu(A);

NN=cell(2);
for i=1:3
    gamma=gamma_vect(i);
    NN{1}=gamma*N;
    NN{2}=gamma*(-N+speye(n));
    starting_block=[C,NN{1}*C,U];
    
    
    % left and right space are the same: use kpik_lyap_LRcomm
    [X1,X2,resnormvec]=EKSM_genLyap_LRcomm(A,LA,UA,NN,C,starting_block,maxit,tol,tolY);
    
    % Do not compute the actual residual norm for very large problems
    relres_true=norm((A*X1)*X2'+X1*(X2'*A')+(NN{1}*X1)*(X2'*NN{1}')+(NN{2}*X1)*(X2'*NN{2}')-C*C','fro')/norm(C*C','fro')

  
    semilogy(resnormvec,'*-')
    xlabel('Iterations')
    ylabel('Relative Residual norm')
    title(['\gamma: ',num2str(gamma)])
    pause
    figure
end
close